import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:graphql_flutter/graphql_flutter.dart' hide JsonSerializable;
import 'package:hooks_riverpod/hooks_riverpod.dart';
import 'package:graphql_hooks_flutter/provider.dart';
import 'package:json_annotation/json_annotation.dart';
import 'package:artemis/schema/graphql_query.dart';
import 'package:artemis/schema/graphql_response.dart';

GraphQLResponse<T> toTyped<T>(QueryResult r, GraphQLQuery<T, dynamic> query) {
  return GraphQLResponse<T>(
      data: r.data == null ? null : query.parse(r.data),
      errors: r.exception == null ? null : r.exception.graphqlErrors);
}

typedef RefetchFunc = Future<QueryResult> Function();
typedef FetchMoreFunc = Future<QueryResult> Function(FetchMoreOptions);

class UseQueryReturn<T, U extends JsonSerializable> {
  ObservableQuery observableQuery;
  GraphQLQuery<T, U> _query;
  GraphQLResponse<T> result;

  Future<GraphQLResponse<T>> refetch() async {
    return this.observableQuery.refetch().then((r) => toTyped(r, _query));
  }

  Future<GraphQLResponse<T>> fetchMore(FetchMoreOptions options) async {
    return this
        .observableQuery
        .fetchMore(options)
        .then((r) => toTyped(r, _query));
  }

  UseQueryReturn(this.result, this._query, this.observableQuery);
}

UseQueryReturn<T, U> useQuery<T, U extends JsonSerializable>(
    GraphQLQuery<T, U> query,
    {Duration pollInterval,
    GraphQLClient graphQLClient,
    StateNotifierProvider<GraphQLClientState> provider}) {
  final client = graphQLClient != null
      ? graphQLClient
      : useProvider(provider != null ? provider : graphQLClientProvider).state;

  final resultState = useState(GraphQLResponse<T>(data: null, errors: null));
  final observableQueryState = useState<ObservableQuery>(null);

  final options = QueryOptions(
      document: query.document,
      operationName: query.operationName,
      variables: query.getVariablesMap(),
      pollInterval: pollInterval);

  useEffect(() {
    final observableQuery = client.watchQuery(options.asWatchQueryOptions());
    observableQueryState.value = observableQuery;

    var stream =
        observableQuery.stream.map((r) => toTyped(r, query)).listen((event) {
      resultState.value = event;
    });

    return () {
      stream.cancel();
      observableQuery.close();
    };
  }, const []);

  return UseQueryReturn(resultState.value, query, observableQueryState.value);
}

typedef RunMutation<U> = MultiSourceResult Function(U variables,
    {Object optimisticResult});

class UseMutationReturn<T, U extends JsonSerializable> {
  ObservableQuery observableMutation;
  GraphQLQuery<T, U> _mutation;
  RunMutation<U> _run;
  GraphQLResponse<T> result;

  Future<GraphQLResponse<T>> run(U variables) async {
    var answer = await this._run(variables).networkResult;
    return toTyped(answer, _mutation);
  }

  UseMutationReturn(
      this._run, this.result, this.observableMutation, this._mutation);
}

UseMutationReturn<T, U> useMutation<T, U extends JsonSerializable>(
    GraphQLQuery<T, U> mutation,
    {Duration pollInterval,
    GraphQLClient graphQLClient,
    OnMutationUpdate update,
    StateNotifierProvider<GraphQLClientState> provider}) {
  final client = graphQLClient != null
      ? graphQLClient
      : useProvider(provider != null ? provider : graphQLClientProvider).state;

  final resultState = useState(GraphQLResponse<T>(data: null, errors: null));
  final observableQueryState = useState<ObservableQuery>(null);

  final options = WatchQueryOptions(
      document: mutation.document,
      operationName: mutation.operationName,
      variables: mutation.getVariablesMap(),
      fetchResults: false);

  final mutation_options = MutationOptions(
      document: mutation.document,
      operationName: mutation.operationName,
      variables: mutation.getVariablesMap(),
      update: update);

  useEffect(() {
    final observableQuery = client.watchMutation(options);
    observableQueryState.value = observableQuery;

    var stream =
        observableQuery.stream.map((r) => toTyped(r, mutation)).listen((event) {
      resultState.value = event;
    });

    return () {
      stream.cancel();
      observableQuery.close();
    };
  }, const []);

  MultiSourceResult runMutation(U variables, {Object optimisticResult}) {
    final mutationCallbacks = MutationCallbackHandler(
        cache: client.cache,
        queryId: observableQueryState.value.queryId,
        options: mutation_options);

    return (observableQueryState.value
          ..variables = variables.toJson()
          ..options.optimisticResult = optimisticResult
          ..onData(mutationCallbacks.callbacks) // add callbacks to observable
        )
        .fetchResults();
  }

  return UseMutationReturn<T, U>(
      runMutation, resultState.value, observableQueryState.value, mutation);
}
